{% extends 'admin/master.html' %}

{% block body %}
  {% if session.user_email %}
    <p>Welcome {{ session.user_email }}!</p>
    <a href="{{ url_for('logout') }}">Logout</a>
    <div>
      <a href="{{ url_for('invoices.index') }}">Invoices</a>
    </div>
  {% else %}
    <a href="{{ url_for('google.login') }}">Login</a>
  {% endif %}
{% endblock %}
